# Super Spelunker
# Copyright (c) 2007 Philip Allison
#
# http://mangobrain.co.uk/
#
# Originally created for the fourth PyWeek Game Contest
# http://www.pyweek.org/4/
#
# Released under the Artistic License - please read LICENSE or License.txt.
# If you did not receive a copy of the license with this file, please contact
# the author via one of the above URLs, stating where you obtained your package.

import pygame
import random
import math

import spResources

#
# Enemy code
#

# Bullets
class Bullet(pygame.sprite.Sprite):
	def __init__(self, x, y, dx, dy, group, globals, room, type = 'bullet'):
		pygame.sprite.Sprite.__init__(self, [group])
		self.image = spResources.getGraphic(globals.gDatadir, type)
		self.rect = pygame.Rect(self.image.get_rect())
		self.rect.centerx = x
		self.rect.centery = y
		self.__dx = dx
		self.__dy = dy
		self.__g = globals
		self.__group = group
		self.__room = room
		self.__fx = float(x)
		self.__fy = float(y)
	def update(self):
		self.__fx += self.__dx
		self.__fy += self.__dy
		self.rect.centerx = int(self.__fx)
		self.rect.centery = int(self.__fy)
		gridx = self.rect.centerx / self.__g.gTileW
		gridy = self.__g.gRoomH - ((self.rect.centery / self.__g.gTileH) - (self.__g.gHudH / self.__g.gTileH)) - 1
		if ((gridx < 0) or (gridx >= self.__g.gRoomW) or (gridy < 0) or (gridy >= self.__g.gRoomH) or (self.__room.grid[gridx][gridy] == 1)):
			self.__group.remove([self])

# Turret enemy
class Turret(pygame.sprite.Sprite):
	def __init__(self, x, y, direction, globals, group, bulletgroup, player, room):
		# Add self to given group
		pygame.sprite.Sprite.__init__(self, [group])
		# Load in sprites according to the direction we've been told to face
		map = {0: 'turret', 1: 'turretl', 2: 'turretd', 3: 'turretr'}
		self.image = spResources.getGraphic(globals.gDatadir, map[direction])
		self.rect = pygame.Rect(0, 0, globals.gTileW, globals.gTileH)
		# Put in the position we've been given
		self.__gridx = x
		self.__gridy = y
		self.rect.move_ip(x * globals.gTileW, globals.gYres - (globals.gTileH * (y+1)))
		# Store reference to bullet group for when we want to fire
		self.__bullets = bulletgroup
		# Load fire animation
		self.__fireanim = spResources.getAnim(globals.gDatadir, map[direction] + '-firing')
		self.__frame = -1
		self.__notfiring = self.image
		# Store reference to player and globals
		self.__player = player
		self.__g = globals
		self.__room = room
		# Not currently exploding
		self.exploding = False
		self.__splodeAnim = spResources.getAnim(globals.gDatadir, 'splode')
		self.__splodeCounter = 0
		# Score value when killed
		self.value = 20

	def update(self):
		# Animate blowing up
		if (self.exploding):
			self.image = self.__splodeAnim[len(self.__splodeAnim) - self.__splodeCounter]
			self.__splodeCounter -= 1
			if (self.__splodeCounter == 0):
				# When finally gone, remove from enemies list
				self.__room.enemies.remove(self)
			return
		# Shoot! Maybe.
		if (random.randint(0, 100) == 1):
			# Calculate distance to player
			dx = self.__player.rect.centerx - self.rect.centerx
			dy = self.__player.rect.centery - self.rect.centery
			# Normalize to a unit vector and multiply by bullet speed
			l = math.sqrt(math.pow(dx, 2) + math.pow(dy, 2))
			dx = (dx/l)*5
			dy = (dy/l)*5
			bullet = Bullet(self.rect.centerx, self.rect.centery, dx, dy, self.__bullets, self.__g, self.__room)
			self.__frame = 15
		if (self.__frame > 10):
			self.image = self.__fireanim[0]
			self.__frame -= 1
		elif (self.__frame > 5):
			self.image = self.__fireanim[1]
			self.__frame -= 1
		elif (self.__frame > 0):
			self.image = self.__fireanim[2]
			self.__frame -= 1
		else:
			self.image = self.__notfiring
	
	def Splode(self):
		if (self.exploding):
			return
		self.exploding = True
		self.__splodeCounter = len(self.__splodeAnim)
		# Mark occupied grid square as empty
		self.__room.grid[self.__gridx][self.__gridy] = 0
		# Undock the player in case they landed on a turret then destroyed it
		self.__player.docked = False

# Laser enemy
class Laser(pygame.sprite.Sprite):
	def __init__(self, x, y, direction, globals, group, player, room):
		# Add self to given group
		pygame.sprite.Sprite.__init__(self, [group])
		# Load in sprites according to the direction we've been told to face
		map = {0: 'laser', 1: 'laserl', 2: 'laserd', 3: 'laserr'}
		self.image = spResources.getGraphic(globals.gDatadir, map[direction])
		self.rect = pygame.Rect(0, 0, globals.gTileW, globals.gTileH)
		self.__direction = direction
		# Put in the position we've been given
		self.__gridx = x
		self.__gridy = y
		self.rect.move_ip(x * globals.gTileW, globals.gYres - (globals.gTileH * (y+1)))
		# Load fire & charge graphics
		self.__fireimage = spResources.getGraphic(globals.gDatadir, map[direction] + '-firing')
		self.__charging = spResources.getAnim(globals.gDatadir, map[direction] + '-charging')
		self.__frame = -1
		self.__notfiring = self.image
		# Store reference to player and globals
		self.__player = player
		self.__g = globals
		self.__room = room
		# Not currently exploding
		self.exploding = False
		self.__splodeAnim = spResources.getAnim(globals.gDatadir, 'splode')
		self.__splodeCounter = 0
		# Not currently shooting
		self.__chargecounter = random.randint(80, 200)
		self.__chargeduration = random.randint(40, 120)
		self.__maxcharge = self.__chargecounter
		self.__maxchargeduration = self.__chargeduration
		self.__linerect = None
		self.__harmcounter = 0
		# Score value when killed
		self.value = 30

	def update(self):
		# Animate blowing up
		if (self.exploding):
			if (self.__linerect != None):
				# Remove laserbeam from screen
				self.__g.gScreen.blit(self.__g.gBackground, self.__linerect, self.__linerect)
				self.__g.gDirtyRects.append(self.__linerect)
				self.__linerect = None
			self.image = self.__splodeAnim[len(self.__splodeAnim) - self.__splodeCounter]
			self.__splodeCounter -= 1
			if (self.__splodeCounter == 0):
				# When finally gone, remove from enemies list
				self.__room.enemies.remove(self)
			return
		# Shoot! Maybe.
		# Charge up laser
		self.__chargecounter -= 1
		if (self.__chargecounter <= 0):
			# Fully charged - shoot
			self.image = self.__fireimage
			# Find endpoint of laser beam (will stop at first non-empty square)
			if (self.__direction == 0):
				px = self.rect.centerx
				py = self.__g.gHudH
				for y in xrange(self.__gridy, self.__g.gRoomH):
					if (self.__room.grid[self.__gridx][y] == 1):
						py = self.__g.gYres - (self.__g.gTileH * y)
						break
			elif (self.__direction == 2):
				px = self.rect.centerx
				py = self.__g.gYres
				for y in xrange(self.__gridy, -1, -1):
					if (self.__room.grid[self.__gridx][y] == 1):
						py = self.__g.gYres - (self.__g.gTileH * (y + 1))
						break
			elif (self.__direction == 1):
				py = self.rect.centery
				px = self.__g.gXres
				for x in xrange(self.__gridx, self.__g.gRoomW):
					if (self.__room.grid[x][self.__gridy] == 1):
						px = self.__g.gTileW * x
						break
			elif (self.__direction == 3):
				py = self.rect.centery
				px = 0
				for x in xrange(self.__gridx, -1, -1):
					if (self.__room.grid[x][self.__gridy] == 1):
						px = self.__g.gTileW * (x + 1)
						break
			self.__chargeduration -= 1
			if (self.__chargeduration == 0):
				# Discharged - stop shooting
				self.image = self.__notfiring
				self.__chargecounter = self.__maxcharge
				self.__chargeduration = self.__maxchargeduration
				# Remove laserbeam from screen
				self.__g.gScreen.blit(self.__g.gBackground, self.__linerect, self.__linerect)
				self.__g.gDirtyRects.append(self.__linerect)
			else:
				# Shoot!
				self.__linerect = pygame.draw.line(self.__g.gScreen, (0, 200, 128), (self.rect.centerx, self.rect.centery), (px, py), 2)
				self.__g.gDirtyRects.append(self.__linerect)
				if ((self.__player.invulnCounter == 0) and (self.__linerect.colliderect(self.__player.rect))):
					# Only harm the player every 3 frames
					if (self.__harmcounter == 0):
						self.__player.shield -= 1
						self.__harmcounter = 2
					else:
						self.__harmcounter -= 1
				else:
					self.__harmcounter = 0
		# Show charge animation
		elif (self.__chargecounter <= 20):
			self.image = self.__charging[1]
		elif (self.__chargecounter <= 40):
			self.image = self.__charging[0]
	
	def Splode(self):
		if (self.exploding):
			return
		self.exploding = True
		self.__splodeCounter = len(self.__splodeAnim)
		# Mark occupied grid square as empty
		self.__room.grid[self.__gridx][self.__gridy] = 0
		# Undock the player in case they landed on a turret then destroyed it
		self.__player.docked = False

# Mine enemy
class Mine(pygame.sprite.Sprite):
	def __init__(self, globals, group):
		pygame.sprite.Sprite.__init__(self, [group])
		# Direction - clockwise or anticlockwise
		self.__direction = random.choice([-3, 3])
		# Starting angle
		self.__angle = random.randint(0, 359)
		# Centre of rotation
		self.__rx = random.randint(100, globals.gXres - 100)
		self.__ry = random.randint(globals.gHudH + 100, globals.gYres - 100)
		# Load graphics
		self.image = spResources.getGraphic(globals.gDatadir, 'mine')
		self.rect = pygame.Rect(0, 0, globals.gTileW, globals.gTileH)
		self.exploding = False
		self.__splodeAnim = spResources.getAnim(globals.gDatadir, 'splode')
		self.__splodeCounter = 0
		self.__group = group

	def update(self):
		# Orbit centre of rotation
		self.__angle += self.__direction
		if (self.__angle < 0):
			self.__angle += 360
		elif (self.__angle >= 360):
			self.__angle -= 360
		r = math.radians(self.__angle)
		x = math.sin(r) * 40
		y = math.cos(r) * 40
		self.rect.centerx = x + self.__rx
		self.rect.centery = y + self.__ry
		# Animate blowing up
		if (self.exploding):
			self.image = self.__splodeAnim[len(self.__splodeAnim) - self.__splodeCounter]
			self.__splodeCounter -= 1
			if (self.__splodeCounter == 0):
				# When finally gone, remove from enemies list & mark grid position as empty
				self.__group.remove(self)
	
	def Splode(self):
		if (self.exploding):
			return
		self.exploding = True
		self.__splodeCounter = len(self.__splodeAnim)

#
# Bosses
#

# Simple explosion sprite class
class Splosion(pygame.sprite.Sprite):
	def __init__(self, x, y, group, globals):
		pygame.sprite.Sprite.__init__(self, [group])
		self.__anim = spResources.getAnim(globals.gDatadir, 'splode')
		self.rect = self.__anim[0].get_rect()
		self.rect.centerx = x
		self.rect.centery = y
		self.__group = group
		self.__counter = 0
	def update(self):
		self.__counter += 1
		if (self.__counter == len(self.__anim)):
			self.__group.remove(self)
			return
		self.image = self.__anim[self.__counter]

class Lobster(pygame.sprite.Sprite):
	def __init__(self, globals, player, group, bulletgroup, room):
		pygame.sprite.Sprite.__init__(self, [group])
		# Load graphic
		self.image = spResources.getGraphic(globals.gDatadir, 'lobster-500')
		self.__painimage = spResources.getGraphic(globals.gDatadir, 'lobster-500-hurt')
		self.__normimage = self.image
		self.rect = self.image.get_rect()
		self.rect.centery = ((globals.gYres - globals.gHudH) / 2) + globals.gHudH
		self.rect.right = globals.gXres - globals.gTileW * 2
		self.health = 50
		self.__g = globals
		self.__direction = 4
		self.exploding = False
		self.__group = group
		self.value = 5000
		self.__hurt = False
		self.__bullets = bulletgroup
		self.__player = player
		self.__room = room
		self.__splodeCounter = 0
		globals.gBossWarning = spResources.getGraphic(globals.gDatadir, 'lobsterwarning');
		globals.gBWTimer = 200
	def update(self):
		if (self.__hurt):
			self.image = self.__painimage
			self.__hurt = False
		else:
			self.image = self.__normimage
		if (self.exploding):
			if (self.__splodeCounter > 0):
				self.__splodeCounter -= 1
				# Spawn explosion graphics
				Splosion(random.randint(self.rect.left, self.rect.right), random.randint(self.rect.top, self.rect.bottom), self.__splodeGroup, self.__g)
			self.__splodeGroup.clear(self.__g.gScreen, self.__g.gBackground)
			self.__splodeGroup.update()
			self.__g.gDirtyRects += self.__splodeGroup.draw(self.__g.gScreen)
			if (self.__splodeCounter % 2 == 0):
				self.image = self.__painimage
			else:
				self.image = self.__normimage
			# Once all explosions have died, remove self from group
			if (len(self.__splodeGroup.sprites()) == 0):
				self.__group.remove(self)
			return
		self.rect.move_ip(0, self.__direction)
		if (self.rect.top < self.__g.gYres - (self.__g.gTileH * (self.__g.gRoomH - 2)) or (self.rect.bottom > self.__g.gYres - (self.__g.gTileH * 2))):
			self.__direction = -self.__direction
		if (self.__player.invulnCounter == 0 and random.randint(0, 5) == 1):
			# Calculate distance to player
			dx = self.__player.rect.centerx - self.rect.centerx
			dy = self.__player.rect.centery - self.rect.centery
			# Normalize to a unit vector and multiply by bullet speed
			l = math.sqrt(math.pow(dx, 2) + math.pow(dy, 2))
			dx = (dx/l)*8
			dy = (dy/l)*8
			bullet = Bullet(self.rect.centerx + random.randint(-10, 10), self.rect.centery + random.randint(-10, 10), dx, dy, self.__bullets, self.__g, self.__room)
			self.__frame = 15
	def Splode(self):
		self.__hurt = True
		if (self.exploding):
			return
		self.health -= 1
		if (self.health < 0):
			self.exploding = True
			self.__splodeCounter = 200
			self.__splodeGroup = pygame.sprite.RenderUpdates()

class Mango(pygame.sprite.Sprite):
	def __init__(self, globals, player, group, bulletgroup, room):
		pygame.sprite.Sprite.__init__(self, [group])
		# Load graphic
		self.image = spResources.getGraphic(globals.gDatadir, 'kitten')
		self.__painimage = spResources.getGraphic(globals.gDatadir, 'kitten-hurt')
		self.__normimage = self.image
		self.rect = self.image.get_rect()
		self.rect.centery = ((globals.gYres - globals.gHudH) / 2) + globals.gHudH
		self.rect.right = globals.gXres - globals.gTileW * 2
		self.health = 80
		self.__g = globals
		self.__xdirection = 6
		self.__ydirection = 6
		self.exploding = False
		self.__group = group
		self.value = 7000
		self.__hurt = False
		self.__bullets = bulletgroup
		self.__player = player
		self.__room = room
		self.__splodeCounter = 0
		globals.gBossWarning = spResources.getGraphic(globals.gDatadir, 'catwarning');
		globals.gBWTimer = 200
	def update(self):
		if (self.__hurt):
			self.image = self.__painimage
			self.__hurt = False
		else:
			self.image = self.__normimage
		if (self.exploding):
			if (self.__splodeCounter > 0):
				self.__splodeCounter -= 1
				# Spawn explosion graphics
				Splosion(random.randint(self.rect.left, self.rect.right), random.randint(self.rect.top, self.rect.bottom), self.__splodeGroup, self.__g)
			self.__splodeGroup.clear(self.__g.gScreen, self.__g.gBackground)
			self.__splodeGroup.update()
			self.__g.gDirtyRects += self.__splodeGroup.draw(self.__g.gScreen)
			if (self.__splodeCounter % 2 == 0):
				self.image = self.__painimage
			else:
				self.image = self.__normimage
			# Once all explosions have died, remove self from group
			if (len(self.__splodeGroup.sprites()) == 0):
				self.__group.remove(self)
			return
		self.rect.move_ip(self.__xdirection, self.__ydirection)
		if (self.rect.top < self.__g.gYres - (self.__g.gTileH * (self.__g.gRoomH - 2)) or (self.rect.bottom > self.__g.gYres - (self.__g.gTileH * 2))):
			self.__ydirection = -self.__ydirection
		if (self.rect.left < (self.__g.gXres / 2) + 30 or self.rect.right > self.__g.gXres - (self.__g.gTileW * 2)):
			self.__xdirection = -self.__xdirection
		if (self.__player.invulnCounter == 0 and random.randint(0, 4) == 1):
			# Calculate distance to player
			dx = self.__player.rect.centerx - self.rect.centerx
			dy = self.__player.rect.centery - self.rect.centery
			# Normalize to a unit vector and multiply by bullet speed
			l = math.sqrt(math.pow(dx, 2) + math.pow(dy, 2))
			dx = (dx/l)*9
			dy = (dy/l)*9
			bullet = Bullet(self.rect.centerx + random.randint(-30, 30), self.rect.centery + random.randint(-30, 30), dx, dy, self.__bullets, self.__g, self.__room)
			self.__frame = 15
	def Splode(self):
		self.__hurt = True
		if (self.exploding):
			return
		self.health -= 1
		if (self.health < 0):
			self.exploding = True
			self.__splodeCounter = 200
			self.__splodeGroup = pygame.sprite.RenderUpdates()
