# Super Spelunker
# Copyright (c) 2007 Philip Allison
#
# http://mangobrain.co.uk/
#
# Originally created for the fourth PyWeek Game Contest
# http://www.pyweek.org/4/
#
# Released under the Artistic License - please read LICENSE or License.txt.
# If you did not receive a copy of the license with this file, please contact
# the author via one of the above URLs, stating where you obtained your package.

import pygame
import random
import os

import spMaze
import spResources
import spEnemies
import spMenu

#
# Game loop code
#

class Game:
	# Start new game/load saved game
	def __init__(self, globals, args):
		# Grab maze grid, distance to furthest point from start, and coordinates of furthest point
		(self.__maze, self.__maxdist, self.__maxX, self.__maxY) = spMaze.GenMaze(globals.gMazeW, globals.gMazeH)
		# Decide locations of keycards
		self.__keys = []
		# Green - make sure it shows up on player's side of green locked rooms
		self.__keys.append([random.randint(0, globals.gMazeW - 1), random.randint(0, globals.gMazeH - 1)])
		while (self.__maze[self.__keys[0][0]][self.__keys[0][1]][5] >= (self.__maxdist / 4)):
			self.__keys[0][1] += 1
			if (self.__keys[0][1] == globals.gMazeH):
				self.__keys[0][1] = 0
				self.__keys[0][0] += 1
				if (self.__keys[0][0] == globals.gMazeW):
					self.__keys[0][0] = 0
		# Blue - make sure the player can get to it, and that it doesn't overlap the green key
		self.__keys.append([random.randint(0, globals.gMazeW - 1), random.randint(0, globals.gMazeH - 1)])
		while ((self.__maze[self.__keys[1][0]][self.__keys[1][1]][5] >= (self.__maxdist / 4) * 2)
			or (self.__keys[1][0] == self.__keys[0][0] and self.__keys[1][1] == self.__keys[0][1])):
			self.__keys[1][1] += 1
			if (self.__keys[1][1] == globals.gMazeH):
				self.__keys[1][1] = 0
				self.__keys[1][0] += 1
				if (self.__keys[1][0] == globals.gMazeW):
					self.__keys[1][0] = 0
		# Red - make sure the player can get to it, and that it doesn't overlap the blue or green keys
		self.__keys.append([random.randint(0, globals.gMazeW - 1), random.randint(0, globals.gMazeH - 1)])
		while ((self.__maze[self.__keys[2][0]][self.__keys[2][1]][5] >= (self.__maxdist / 4) * 3)
			or (self.__keys[2][0] == self.__keys[0][0] and self.__keys[2][1] == self.__keys[0][1])
			or (self.__keys[2][0] == self.__keys[1][0] and self.__keys[2][1] == self.__keys[1][1])):
			self.__keys[2][1] += 1
			if (self.__keys[2][1] == globals.gMazeH):
				self.__keys[2][1] = 0
				self.__keys[2][0] += 1
				if (self.__keys[2][0] == globals.gMazeW):
					self.__keys[2][0] = 0
		# Pick a random point to be the exit (always on y == globals.gMazeH - 1, so just pick an x)
		# Don't pick the artifact chamber (furthest point from start)
		self.__exitX = random.randint(0, globals.gMazeW - 1)
		while ((self.__exitX == self.__maxX) and (self.__maxY == globals.gMazeH - 1)):
			self.__exitX += 1
			if (self.__exitX == globals.gMazeW):
				self.__exitX = 0
		# Add Room object references, so a room can be generated when first entered, then saved
		# Initially, no rooms have been generated
		self.__rooms = [[None for iy in xrange(globals.gMazeH)] for ix in xrange(globals.gMazeW)]
		# Create player
		self.__player = Player(globals)
		# Player's position in maze coordinates (centre column, bottom row)
		self.__pmazex = globals.gMazeW/2
		self.__pmazey = 0
		# Player enters the first from in starting position
		# Possible room entrances: 0 - left, 1 - right, 2 - top, 3 - bottom, 4 - start
		self.__entrance = 4
		self.__player.enterRoom(self.__entrance, False)
		# Track bullets in current room
		self.__bullets = pygame.sprite.RenderUpdates()
		self.__playerbullets = pygame.sprite.RenderUpdates()
		# Generate first room
		self.__rooms[self.__pmazex][self.__pmazey] = spMaze.Room(globals, self.__maze, self.__pmazex, self.__pmazey, self.__bullets, self.__player, self.__keys, True)
		self.__room = self.__rooms[self.__pmazex][self.__pmazey]
		self.__background = self.__room.drawBackground(globals)
		globals.gBackground = self.__background
		# Set the game parameters
		# Gravity strength
		globals.gGravityY = 0.20
		globals.gGravityX = 0
		# Screen edges
		# left, right, top, bottom
		self.__minx = 0
		self.__maxx = globals.gXres
		self.__miny = globals.gYres - (globals.gTileH * globals.gRoomH)
		self.__maxy = globals.gYres - globals.gTileH
		# Store initial state of HUD info and draw the HUD
		self.__score = 0
		self.__oldscore = 0
		self.__oldshield = self.__player.shield
		self.__oldlives = self.__player.lives
		# Draw initial game screen
		globals.gScreen.blit(self.__background, (0, 0))
		self.updateHud(globals, True)
		globals.gDirtyRects.append(pygame.Rect(0, 0, globals.gXres, globals.gYres))
		# Not currently showing map
		self.__showingMap = False
		self.__map = None
		self.__paused = False
	
	# Update the HUD - either forcibly, or only as necessary
	def updateHud(self, globals, force = False):
		changed = False
		if (force or (self.__score != self.__oldscore)):
			self.__oldscore = self.__score
			font = spResources.getDefaultFont()
			text = font.render(str(self.__score), True, (255, 128, 128), (0, 0, 0))
			rect = text.get_rect()
			rect.centerx = 700
			rect.centery = 65
			globals.gScreen.blit(text, rect)
			changed = True
		if (force or (self.__player.lives != self.__oldlives)):
			self.__oldlives = self.__player.lives
			font = spResources.getDefaultFont()
			text = font.render(str(self.__player.lives), True, (255, 128, 128), (0, 0, 0))
			rect = text.get_rect()
			rect.centerx = 450
			rect.centery = 65
			globals.gScreen.blit(text, rect)
			changed = True
		if (force or (self.__player.shield != self.__oldshield)):
			changed = True
			self.__oldshield = self.__player.shield
			shieldslice = spResources.getGraphic(globals.gDatadir, 'shieldslice', False)
			depletedshield = spResources.getGraphic(globals.gDatadir, 'depletedshieldslice', False)
			for x in xrange(0, globals.gMaxShield):
				rect = shieldslice.get_rect()
				rect.left = 395 - (x*5)
				rect.centery = 65
				if (x < self.__player.shield):
					globals.gScreen.blit(shieldslice, rect)
				else:
					globals.gScreen.blit(depletedshield, rect)
		if (0 in self.__player.heldKeys):
			globals.gScreen.blit(spResources.getGraphic(globals.gDatadir, 'greenkeycard'), (510, 55))
		if (1 in self.__player.heldKeys):
			globals.gScreen.blit(spResources.getGraphic(globals.gDatadir, 'bluekeycard'), (530, 55))
		if (2 in self.__player.heldKeys):
			globals.gScreen.blit(spResources.getGraphic(globals.gDatadir, 'redkeycard'), (550, 55))
		if (3 in self.__player.heldKeys):
			globals.gScreen.blit(spResources.getGraphic(globals.gDatadir, 'artifact'), (570, 55))
		if (force or changed):
			globals.gDirtyRects.append(pygame.Rect(0, 0, globals.gXres, globals.gHudH))
	
	# Run once each time round the main loop - perform high-level game logic
	def tick(self, globals):
		# Game over?
		if (self.__player.lives < 0):
			globals.gNextLoopClass = spMenu.GameOver
			return (True, (False, self.__score)) 
		kp = pygame.event.get(pygame.KEYDOWN)
		# Toggle pause/unpause
		togglep = False
		# Show boss warnings OR allow pausing - too confusing trying to interleave the two
		# in the current rat's nest of code.
		if (globals.gBossWarning != None):
			if (globals.gBWTimer == 200):
				globals.gScreen.blit(globals.gBossWarning, (0, 0))
				globals.gDirtyRects.append(pygame.Rect(0, 0, globals.gXres, globals.gYres))
			globals.gBWTimer -= 1
			if (globals.gBWTimer == 0):
				globals.gBossWarning = None
				togglep = True
			else:
				return (False, None)
		else:
			for k in (kp):
				if (k.key == pygame.K_p):
					self.__paused = not self.__paused
					togglep = True
			if (self.__paused):
				if (togglep):
					p = spResources.getGraphic(globals.gDatadir, 'paused')
					globals.gScreen.blit(p, (0, 0))
					globals.gDirtyRects.append(pygame.Rect(0, 0, globals.gXres, globals.gYres))
				return (False, None)
		# Wipe map, paused screen or boss warning if was being shown
		if (self.__showingMap or togglep):
			self.__showingMap = (globals.gKeyState[pygame.K_SPACE] and not togglep)
			globals.gScreen.blit(self.__background, (0, 0))
			self.updateHud(globals, True)
			globals.gDirtyRects.append(pygame.Rect(0, 0, globals.gXres, globals.gYres))
		# Wipe old player position
		globals.gScreen.blit(self.__background, self.__player.rect, self.__player.rect)
		# Set current player rect as dirty - instantiate a new copy, since Python will
		# use references, and hence both "old" and "new" locations will always be identical
		# TODO: Clip against play area?
		globals.gDirtyRects.append(pygame.Rect(self.__player.rect))

		# Move to new room if off screen edge
		# Calculate left and top of ship in grid coordinates here; use those to determine
		# whether or not we're out of screen such that we can't end up
		# failing to move then trying to collide with out-of-range grid squares
		left_x = self.__player.rect.left / globals.gTileW
		top_y = (globals.gRoomH - ((self.__player.rect.top / globals.gTileH) - (globals.gHudH / globals.gTileH)) - 1)
		moved = False
		if (left_x < 0):
			moved = True
			self.__entrance = 1
		elif (left_x + 1 >= globals.gRoomW):
			moved = True
			self.__entrance = 0
		elif (top_y >= globals.gRoomH):
			moved = True
			self.__entrance = 3
		elif (top_y - 1 < 0):
			moved = True
			self.__entrance = 2

		if (not moved):
			# Collide player against static environment elements (use spritegroups for mobiles)
			# Check the four grid squares we're currently spanning
			# Y coordinates must take into account billboard at top of screen, and reversed screen Y axis (relative to grid Y)
			# Blow up if collided with left, right or top sides; call Player.land if collided with bottom side
			# to decide whether to land (dock) or blow up.
			# Colliding with bottom is defined by bottom left/right square being filled and player moving downwards
			if (not self.__player.docked):
				if (self.__room.squareIsSolid(left_x, top_y) or self.__room.squareIsSolid(left_x + 1, top_y)):
					# Bounce, land or explode, depending upon context & shield levels
					self.__player.Bounce(self.__entrance, self.__room)
				elif (self.__room.squareIsSolid(left_x, top_y - 1) or self.__room.squareIsSolid(left_x + 1, top_y - 1)):
					if (self.__player.vy >= 0):
						if (not self.__player.Land(top_y)):
							self.__player.Bounce(self.__entrance, self.__room)
					else:
						self.__player.Bounce(self.__entrance, self.__room)

			# Collide player against mobiles (bullets, bats, icicles)
			if ((globals.gGameMode == 0) and (self.__player.invulnCounter == 0)):
				self.__player.shield -= len(pygame.sprite.spritecollide(self.__player, self.__bullets, True)) * 2
				for enemy in (pygame.sprite.spritecollide(self.__player, self.__room.mobile_enemies, False)):
					if (not enemy.exploding):
						self.__player.shield -= 5
						enemy.Splode()
				if (self.__player.shield < 0):
					self.__player.enterRoom(self.__entrance, self.__room.bossRoom, True)

			# Apply player controls to ship - includes velocity clamping
			self.__player.update()

			# If in adventure mode...
			if (globals.gGameMode == 0):
				# Let turrets shoot, etc.
				self.__room.mobile_enemies.clear(globals.gScreen, self.__background)
				self.__room.mobile_enemies.update()
				if (self.__room.bossRoom):
					self.__room.boss_enemies.clear(globals.gScreen, self.__background)
					self.__room.boss_enemies.update()
				self.__room.enemies.clear(globals.gScreen, self.__background)
				self.__room.enemies.update()
				# Erase old bullet positions
				self.__bullets.clear(globals.gScreen, self.__background)
				self.__bullets.update()
				# Allow player to shoot
				if (self.__player.splodeCounter == 0):
					for key in (kp):
						if (key.key == pygame.K_w):
							spEnemies.Bullet(self.__player.rect.centerx, self.__player.rect.centery, 0, -18, self.__playerbullets, globals, self.__room, 'player-bullet')
						if (key.key == pygame.K_a):
							spEnemies.Bullet(self.__player.rect.centerx, self.__player.rect.centery, -18, 0, self.__playerbullets, globals, self.__room, 'player-bullet')
						if (key.key == pygame.K_d):
							spEnemies.Bullet(self.__player.rect.centerx, self.__player.rect.centery, 18, 0, self.__playerbullets, globals, self.__room, 'player-bullet')
						if (key.key == pygame.K_s):
							spEnemies.Bullet(self.__player.rect.centerx, self.__player.rect.centery, 0, 18, self.__playerbullets, globals, self.__room, 'player-bullet')
				# Collide player bullets against enemies and harm them
				self.__playerbullets.clear(globals.gScreen, self.__background)
				self.__playerbullets.update()
				collisions = pygame.sprite.groupcollide(self.__playerbullets, self.__room.enemies, False, False)
				for bullet in (collisions.iterkeys()):
					enemies = collisions[bullet]
					for enemy in enemies:
						if (not enemy.exploding):
							enemy.Splode()
							# We killed something; increase our score
							self.__score += enemy.value
							self.__playerbullets.remove(bullet)
				if (self.__room.bossRoom):
					# Collide player bullets against bosses
					collisions = pygame.sprite.groupcollide(self.__playerbullets, self.__room.boss_enemies, False, False)
					for bullet in (collisions.iterkeys()):
						enemies = collisions[bullet]
						for enemy in enemies:
							px = bullet.rect.centerx - enemy.rect.left
							py = bullet.rect.centery - enemy.rect.top
							# Bosses are not likely to be rectangular - use per-pixel collision detection
							if ((-1 < px < enemy.rect.width)
								and (-1 < py < enemy.rect.height)
								and (enemy.image.get_at((px, py)) != enemy.image.get_colorkey())):
								if (not enemy.exploding):
									enemy.Splode()
									if (enemy.health < 0):
										self.__score += enemy.value
									self.__playerbullets.remove(bullet)
					# This room ain't got no boss no more
					if (len(self.__room.boss_enemies.sprites()) == 0):
						self.__room.bossRoom = False
						self.__player.heldKeys.append(self.__room.whichKey)
						self.__background = self.__room.drawBackground(globals)
						globals.gBackground = self.__background
						globals.gScreen.blit(self.__background, (0, 0))
						self.updateHud(globals, True)
						globals.gDirtyRects = []
						globals.gDirtyRects.append(pygame.Rect(0, 0, globals.gXres, globals.gYres))
						self.__showingMap = False
					# Collide ship against bosses
					if (self.__player.invulnCounter == 0):
						collisions = pygame.sprite.spritecollide(self.__player, self.__room.boss_enemies, False)
						for boss in collisions:
							px = self.__player.rect.centerx - boss.rect.left
							py = self.__player.rect.centery - boss.rect.top
							if ((-1 < px < boss.rect.width)
								and (-1 < py < boss.rect.height)
								and (boss.image.get_at((px, py)) != boss.image.get_colorkey())):
								if (not boss.exploding):
									self.__player.shield = -1

				# Draw bullets and enemies in new positions
				globals.gDirtyRects += self.__room.enemies.draw(globals.gScreen)
				globals.gDirtyRects += self.__bullets.draw(globals.gScreen)
				globals.gDirtyRects += self.__room.mobile_enemies.draw(globals.gScreen)
				if (self.__room.bossRoom):
					globals.gDirtyRects += self.__room.boss_enemies.draw(globals.gScreen)
				globals.gDirtyRects += self.__playerbullets.draw(globals.gScreen)

		else:
			# Player has changed rooms
			if (globals.gGameMode == 0):
				self.__bullets.empty()
				self.__playerbullets.empty()
			if (self.__entrance == 1):
				self.__pmazex -= 1
			elif (self.__entrance == 0):
				self.__pmazex += 1
			elif (self.__entrance == 3):
				self.__pmazey += 1
			else:
				self.__pmazey -= 1
			# Game complete?
			if (self.__pmazex == self.__exitX and self.__pmazeY == globals.gMazeH - 1 and 3 in self.__player.heldKeys):
				globals.gNextLoopClass = spMenu.GameOver
				return (True, (True, self.__score)) 
			# Generate current room if necessary
			self.__room = self.__rooms[self.__pmazex][self.__pmazey]
			if (self.__room == None):
				# Create room with walls sealed according to maze
				self.__rooms[self.__pmazex][self.__pmazey] = spMaze.Room(globals, self.__maze, self.__pmazex, self.__pmazey, self.__bullets, self.__player, self.__keys)
				self.__room = self.__rooms[self.__pmazex][self.__pmazey]
				# Score points for finding an unexplored room
				self.__score += 100
			# Put player in starting pos
			self.__player.enterRoom(self.__entrance, self.__room.bossRoom)
			self.__background = self.__room.drawBackground(globals)
			globals.gBackground = self.__background
			globals.gScreen.blit(self.__background, (0, 0))
			globals.gDirtyRects = []
			globals.gDirtyRects.append(pygame.Rect(0, 0, globals.gXres, globals.gYres))

		# Draw player sprite & add its position to the list of dirty rects
		globals.gScreen.blit(self.__player.image, self.__player.rect)
		# Update HUD
		self.updateHud(globals, moved)
		# TODO: Clip against play area?
		globals.gDirtyRects.append(self.__player.rect)

		# Show map
		if (globals.gKeyState[pygame.K_SPACE]):
			if (moved or not self.__showingMap):
				self.__map = spMaze.DrawMaze(globals, self.__maze, self.__rooms, self.__pmazex, self.__pmazey, self.__exitX, self.__keys, self.__player)
			minimap = globals.gScreen.copy()
			globals.gScreen.blit(self.__map, (0, 0))
			globals.gScreen.blit(minimap, (10, 450), pygame.Rect(self.__player.rect.centerx - 70, self.__player.rect.centery - 70, 140, 140))
			globals.gDirtyRects.append(pygame.Rect(0, 0, globals.gXres, globals.gYres))
			self.__showingMap = True

		return (False, None)

#
# Player code
#

class Player(pygame.sprite.Sprite):

	def __init__(self, globals, groups = []):
		# Call base initialiser
		pygame.sprite.Sprite.__init__(self, groups)
		# Get reference to player ship graphic
		self.image = spResources.getGraphic(globals.gDatadir, 'player')
		# Set width & height
		self.rect = pygame.Rect(0, 0, globals.gTileW, globals.gTileH)
		# Keep a reference to the globals record
		self.__g = globals
		# Set default velocity
		self.vx = 0
		self.vy = 0
		# Set max/min velocities
		self.__mx = globals.gTileW/2
		self.__my = globals.gTileH/2
		# Not docked to start with
		self.docked = False
		# We have 100% shield
		self.shield = globals.gMaxShield
		# We have no keys
		self.heldKeys = []
		# We have just spawned
		self.invulnCounter = globals.gInvulnFrames
		self.splodeCounter = 0
		self.__invulnImage = spResources.getGraphic(globals.gDatadir, 'invulnplayer')
		self.__playerImage = self.image
		self.__splodeAnim = spResources.getAnim(globals.gDatadir, 'player-splode')
		self.__oldPos = None
		self.__newPos = self.rect
		self.lives = 5

	def update(self):
		# Draw explosion if sploding
		if (self.splodeCounter > 0):
			self.image = self.__splodeAnim[(len(self.__splodeAnim) - self.splodeCounter) - 1]
			self.splodeCounter -= 1
			if (self.splodeCounter == len(self.__splodeAnim)):
				self.__newPos = self.rect
			self.rect = self.__oldPos
			return
		# Flash & remain stationary if invulnerable
		if (self.invulnCounter > 0):
			self.rect = self.__newPos
			if (self.invulnCounter % 2 == 0):
				self.image = self.__invulnImage
			else:
				self.image = self.__playerImage
			self.invulnCounter -= 1
			self.docked = False
			return
		self.image = self.__playerImage
		# Move ship according to environment & controls pressed
		if (not self.docked):
			if (self.__g.gKeyState[pygame.K_LEFT]):
				self.vx -= 0.5
			elif (self.__g.gKeyState[pygame.K_RIGHT]):
				self.vx += 0.5
			self.vy += self.__g.gGravityY
			self.vx += self.__g.gGravityX
		if ((not self.docked) and self.__g.gKeyState[pygame.K_DOWN]):
			self.vy -= 0.25
		elif (self.__g.gKeyState[pygame.K_UP]):
			self.vy -= 0.70
			# Give player chance to get up to speed before
			# declaring undocked, or you get into infinite
			# docking/undocking loop
			if (self.vy < -1):
				self.docked = False
		# Clamp to maximum velocities
		if (self.vx < -self.__mx):
			self.vx = -self.__mx
		elif (self.vx > self.__mx):
			self.vx = self.__mx
		if (self.vy < -self.__my):
			self.vy = -self.__my
		elif (self.vy > self.__my):
			self.vy = self.__my
		# Perform move
		self.rect.move_ip(self.vx, self.vy)
	
	def enterRoom(self, entrance, bossRoom, stop = False):
		# Put player in sensible screen-space position, based on where they came from
		# If the player just died, set both their coordinates, so we can guarantee they're in empty space
		# Always leave some padding so as not to trigger infinite loop of room entrance/exit
		if (stop):
			self.__oldPos = pygame.Rect(self.rect)
		if (bossRoom):
			# boss room - enter in middle and force stop
			self.rect.centerx = self.__g.gXres / 2
			self.rect.centery = ((self.__g.gYres - self.__g.gHudH) / 2) + self.__g.gHudH
			self.vx = 0
			self.vy = 0
			self.invulnCounter = self.__g.gInvulnFrames
		elif (entrance == 0):
			# from the left
			self.rect.left = 5
			if (stop):
				self.rect.top = self.__g.gYres - (self.__g.gTileH * (self.__g.gRoomH / 2))
		elif (entrance == 1):
			# from the right
			self.rect.right = self.__g.gXres - 5
			if (stop):
				self.rect.top = self.__g.gYres - (self.__g.gTileH * (self.__g.gRoomH / 2))
		elif (entrance == 2):
			# from the top
			if (stop):
				self.rect.left = self.__g.gTileW * (self.__g.gRoomW / 2)
			self.rect.top = (self.__g.gYres - (self.__g.gTileH * self.__g.gRoomH)) + 5
		elif (entrance == 3):
			# from the bottom
			if (stop):
				self.rect.left = self.__g.gTileW * (self.__g.gRoomW / 2)
			self.rect.bottom = self.__g.gYres - 5
		else:
			# initial starting position in room 1
			self.rect.left = (self.__g.gRoomW / 2) * self.__g.gTileW
			self.rect.bottom = self.__g.gYres - self.__g.gTileH
		if (stop):
			self.vy = 0
			self.vx = 0
			self.splodeCounter = len(self.__splodeAnim)
			self.lives -= 1
			self.invulnCounter = self.__g.gInvulnFrames
			self.shield = self.__g.gMaxShield
	
	def Land(self, ship_top):
		# Can only land if not going too fast left/right/down
		if ((-2 < self.vx < 2) and (self.vy < 4)):
			self.vx = 0
			self.vy = 0
			self.docked = True
			self.rect.bottom = self.__g.gYres - (ship_top * self.__g.gTileH)
			return True
		return False
	
	def Bounce(self, entrance, room):
		# Take off some energy, and re-spawn if necessary
		if (not self.invulnCounter > 0):
			self.shield -= 1
		if (self.shield < 0):
			self.enterRoom(entrance, room.bossRoom, True)

		# Move to old position to resolve the initial collision
		self.rect.move_ip(-self.vx, -self.vy)
		# Move by vx and see if it triggers a collision
		testrect = pygame.Rect(self.rect)
		testrect.move_ip(self.vx, 0)
		left_x = testrect.left / self.__g.gTileW
		top_y = (self.__g.gRoomH - ((testrect.top / self.__g.gTileH) - (self.__g.gHudH / self.__g.gTileH)) - 1)
		if (room.squareIsSolid(left_x, top_y) or room.squareIsSolid(left_x, top_y - 1) or room.squareIsSolid(left_x + 1, top_y) or room.squareIsSolid(left_x + 1, top_y - 1)):
			self.vx = -self.vx
		# Try the Y direction
		testrect = pygame.Rect(self.rect)
		testrect.move_ip(self.vx, self.vy)
		left_x = testrect.left / self.__g.gTileW
		top_y = (self.__g.gRoomH - ((testrect.top / self.__g.gTileH) - (self.__g.gHudH / self.__g.gTileH)) - 1)
		if (room.squareIsSolid(left_x, top_y) or room.squareIsSolid(left_x, top_y - 1) or room.squareIsSolid(left_x + 1, top_y) or room.squareIsSolid(left_x + 1, top_y - 1)):
			self.vy = -self.vy
