# Super Spelunker
# Copyright (c) 2007 Philip Allison
#
# http://mangobrain.co.uk/
#
# Originally created for the fourth PyWeek Game Contest
# http://www.pyweek.org/4/
#
# Released under the Artistic License - please read LICENSE or License.txt.
# If you did not receive a copy of the license with this file, please contact
# the author via one of the above URLs, stating where you obtained your package.

import pygame
import random
import sys

import spResources
import spEnemies

#
# Maze functions
#

# Maze generation algorithm
# Specify maze size
def GenMaze(x, y):
	global maxx, maxy, maxdist
	# Start off with a blank maze
	# Flags: visited, upper wall, lower wall, left wall, right wall, distance travelled to get here
	maze = [[[0, 1, 1, 1, 1, sys.maxint] for iy in xrange(y)] for ix in xrange(x)]
	# Insert the start point, to which you have had to travel 0 distance
	stack = [(x/2, 0, 0)]
	maze[x/2][0][0] = 1
	# Keep track of the square which is furthest away from the player
	maxdist = 0
	maxx = x/2
	maxy = 0
	# Map of deltas to neighbour squares (up, down, left, right)
	neighbours = ((0, 1), (0, -1), (-1, 0), (1, 0))
	# Map of opposite walls (up -> down, left -> right, etc.)
	opposites = (1, 0, 3, 2)
	# Weighted roulette wheel for picking neighbours - make it more likely we go left/right than up/down
	weights = (2, 2, 2, 3, 3, 3, 0, 1)
	# Keep going until the stack of squares with unvisited neighbours is empty
	while (len(stack) > 0):
		# Grab a point from the stack
		# Occasionally, pick from somewhere other than the top of the stack,
		# to increase the amount of branching
		if (random.randint(0, 7) == 7):
			index = random.randint(0, len(stack) - 1)
		else:
			index = len(stack) - 1
		(currx, curry, currdist) = stack[index]
		# Pick a neighbour "at random", using the weighted roulette wheel for bias
		# If we don't do this, the pyramid layout of the maze gves a natural bias
		# towards long vertical corridors
		neighbour = weights[random.randint(0, 7)]
		# Flag to say whether or not this square was a dead end
		deadend = True
		# Check this neighbour first, but iterate over all others if necessary
		first = True
		for n in [neighbour] + range(3):
			nx = currx + neighbours[n][0]
			ny = curry + neighbours[n][1]
			# Apply pyramid shape constraint
			lowxlimit = (x/2) - ny
			if (lowxlimit < 0):
				lowxlimit = 0
			highxlimit = (x/2) + ny + 1
			if (highxlimit > x):
				highxlimit = x
			# Neighbour coordinates must be in range;
			# neighbour must not already be visited
			if ((lowxlimit <= nx < highxlimit) and (0 <= ny < y) and (maze[nx][ny][0] == 0)):
				# Criteria matched! Knock down walls between here and neighbour
				deadend = False
				# Knock down one wall from current square, and opposite wall in neighbour square
				maze[currx][curry][n + 1] = 0
				maze[nx][ny][opposites[n] + 1] = 0
				# Mark neighbour as visited
				maze[nx][ny][0] = 1
				# Mark neighbour as currdist+1 away from start
				maze[nx][ny][5] = currdist + 1
				# Add new neighbour to stack
				stack.append((nx, ny, currdist + 1))
				if ((currdist + 1) > maxdist):
					maxdist = currdist + 1
					maxx = nx
					maxy = ny
				# Pick only one neighbour per iteration
				break
		# Remove from stack if dead end
		if (deadend):
			# Remove by concatenating slices taken above/below current index,
			# since this isn't always strictly used as a stack.
			stack = stack[0:index] + stack[index+1:len(stack)]
	return (maze, maxdist, maxx, maxy)

# Maze map drawing function
def DrawMaze(globals, maze, rooms, playerx, playery, exitx, keys, player):
	global maxx, maxy, maxdist
	surf = pygame.Surface((globals.gXres, globals.gYres))
	surf.fill((90, 50, 12))
	surf.convert()
	xdelta = (globals.gXres / globals.gMazeW)
	ydelta = (globals.gYres / globals.gMazeH)
	x = 0
	gx = 0
	for column in maze:
		gy = 0
		y = globals.gYres - 1
		for row in column:
			if (row[1] == 1 and row[2] == 1 and row[3] == 1 and row[4] == 1):
				# fill square in if all walls present
				pygame.draw.rect(surf, (145, 96, 53), pygame.Rect(x, y - ydelta, xdelta, ydelta))
			else:
				# highlight player's current location
				if (gx == playerx and gy == playery):
					pygame.draw.rect(surf, (255, 128, 128), pygame.Rect(x, y - ydelta, xdelta, ydelta))
				# highlight furthest room unless player holds artifact
				elif (gx == maxx and gy == maxy and not (3 in player.heldKeys)):
					pygame.draw.rect(surf, (255, 255, 0), pygame.Rect(x, y - ydelta, xdelta, ydelta))
				# highlight red, green and blue keys unless already retrieved
				if (gx == keys[0][0] and gy == keys[0][1] and not (0 in player.heldKeys)):
					pygame.draw.rect(surf, (0, 255, 0), pygame.Rect(x, y - ydelta, xdelta, ydelta))
				elif (gx == keys[1][0] and gy == keys[1][1] and not (1 in player.heldKeys)):
					pygame.draw.rect(surf, (0, 0, 255), pygame.Rect(x, y - ydelta, xdelta, ydelta))
				elif (gx == keys[2][0] and gy == keys[2][1] and not (2 in player.heldKeys)):
					pygame.draw.rect(surf, (255, 0, 0), pygame.Rect(x, y - ydelta, xdelta, ydelta))
				# highlight exit
				if (gy == globals.gMazeH - 1 and gx == exitx):
					pygame.draw.rect(surf, (255, 255, 255), pygame.Rect(x, y - ydelta, xdelta, ydelta))
				if (row[1] == 1):
					# draw upper wall
					pygame.draw.line(surf, (145, 96, 53), (x, y - ydelta), (x + xdelta, y - ydelta), 2)
				if (row[2] == 1):
					# draw lower wall
					pygame.draw.line(surf, (145, 96, 53), (x, y), (x + xdelta, y), 2)
				if (row[3] == 1):
					# draw left wall
					pygame.draw.line(surf, (145, 96, 53), (x, y), (x, y - ydelta), 2)
				if (row[4] == 1):
					# draw right wall
					pygame.draw.line(surf, (145, 96, 53), (x + xdelta, y), (x + xdelta, y - ydelta), 2)
				# draw barriers around locked rooms
				if (row[5] == (maxdist / 4) and not (0 in player.heldKeys)):
					pygame.draw.rect(surf, (0, 255, 0), pygame.Rect(x, y - ydelta, xdelta, ydelta), 1)
				elif (row[5] == (maxdist / 4) * 2 and not (1 in player.heldKeys)):
					pygame.draw.rect(surf, (0, 0, 255), pygame.Rect(x, y - ydelta, xdelta, ydelta), 1)
				elif (row[5] == (maxdist / 4) * 3 and not (2 in player.heldKeys)):
					pygame.draw.rect(surf, (255, 0, 0), pygame.Rect(x, y - ydelta, xdelta, ydelta), 1)
			y -= ydelta
			gy += 1
		x += xdelta
		gx += 1
	# Put outline around miniview location
	pygame.draw.rect(surf, (128, 0, 180), (0, 440, 160, 160))
	pygame.draw.rect(surf, (0, 0, 0), (10, 450, 140, 140))
	# Add legend
	legend = spResources.getGraphic(globals.gDatadir, 'legend', False)
	surf.blit(legend, (640, 440))
	return surf

#
# Maze room population code
#

class Room:
	def __init__(self, globals, maze, mazeX, mazeY, bulletgroup, player, keys, startRoom = False):
		global maxdist, maxx, maxy
		self.__player = player
		# Give artifact to player if in furthest room - running out of time to make a proper boss!
		# TODO: Implement something proper here
		if (mazeX == maxx and mazeY == maxy):
			self.__player.heldKeys.append(3)
		# Create room array
		self.grid = [[0 for iy in xrange(globals.gRoomH)] for ix in xrange(globals.gRoomW)]
		if (globals.gGameMode == 0):
			# Check to see if we're next to a locked room
			# The room furthest away isn't locked, but its [5] *will* be divisible by maxdist/4
			if (maze[mazeX][mazeY][5] < maxdist - 1):
				if ((mazeY < globals.gMazeH - 1) and (maze[mazeX][mazeY + 1][5] % (maxdist / 4) == 0)):
					# Locked door to upper room
					colour = (maze[mazeX][mazeY + 1][5] / (maxdist / 4)) + 1
					if ((2 <= colour <= 4) and not ((colour - 2) in player.heldKeys)):
						for x in xrange(0, globals.gRoomW):
							self.grid[x][globals.gRoomH - 1] = colour
				if ((mazeY > 0) and (maze[mazeX][mazeY - 1][5] % (maxdist / 4) == 0)):
					# Locked door to lower room
					colour = (maze[mazeX][mazeY - 1][5] / (maxdist / 4)) + 1
					if ((2 <= colour <= 4) and not ((colour - 2) in player.heldKeys)):
						for x in xrange(0, globals.gRoomW):
							self.grid[x][0] = colour
				if ((mazeX < globals.gMazeW - 1) and (maze[mazeX + 1][mazeY][5] % (maxdist / 4) == 0)):
					# Locked door to right room room
					colour = (maze[mazeX + 1][mazeY][5] / (maxdist / 4)) + 1
					if ((2 <= colour <= 4) and not ((colour - 2) in player.heldKeys)):
						for y in xrange(0, globals.gRoomH):
							self.grid[globals.gRoomW - 1][y] = colour
				if ((mazeY > 0) and (maze[mazeX - 1][mazeY][5] % (maxdist / 4) == 0)):
					# Locked door to left room
					colour = (maze[mazeX - 1][mazeY][5] / (maxdist / 4)) + 1
					if ((2 <= colour <= 4) and not ((colour - 2) in player.heldKeys)):
						for y in xrange(0, globals.gRoomH):
							self.grid[0][y] = colour
			# Is this a boss (keycard) room?
			self.bossRoom = False
			self.whichKey = 0
			for key in (keys):
				if (mazeX == key[0] and mazeY == key[1]):
					self.bossRoom = True
					break
				self.whichKey += 1
		else:
			self.bossRoom = False
		# Fill in blocked walls
		# Ensure walls are only 2 blocks thick at each end, to ensure no invisible outcrops
		# are present when switching rooms, but allow varying thickness in between
		# Upper
		if (maze[mazeX][mazeY][1] == 1):
			# Starting thickness
			wallThickness = 2
			# Allow thickness to vary up to 4 blocks away from opposite end.
			range1 = xrange(globals.gRoomW-4)
			range2 = xrange(globals.gRoomW-4, globals.gRoomW)
			for x in range1:
				for y in xrange(wallThickness):
					self.grid[x][(globals.gRoomH - 1) - y] = 1
				wallThickness = self.__modifyWallWidth(wallThickness)
			# As we approach opposite end, converge thickness back towards 2
			for x in range2:
				for y in xrange(wallThickness):
					self.grid[x][(globals.gRoomH - 1) - y] = 1
				if (wallThickness > 2):
					wallThickness -= 1
				elif (wallThickness < 2):
					wallThickness += 1
		# Lower
		if (maze[mazeX][mazeY][2] == 1):
			wallThickness = 2
			range1 = xrange(globals.gRoomW-4)
			range2 = xrange(globals.gRoomW-4, globals.gRoomW)
			for x in range1:
				for y in xrange(wallThickness):
					self.grid[x][y] = 1
				wallThickness = self.__modifyWallWidth(wallThickness)
			for x in range2:
				for y in xrange(wallThickness):
					self.grid[x][y] = 1
				if (wallThickness > 2):
					wallThickness -= 1
				elif (wallThickness < 2):
					wallThickness += 1
		# Left
		if (maze[mazeX][mazeY][3] == 1):
			wallThickness = 2
			range1 = xrange(globals.gRoomH-4)
			range2 = xrange(globals.gRoomH-4, globals.gRoomH)
			for y in range1:
				for x in xrange(wallThickness):
					self.grid[x][y] = 1
				wallThickness = self.__modifyWallWidth(wallThickness)
			for y in range2:
				for x in xrange(wallThickness):
					self.grid[x][y] = 1
				if (wallThickness > 2):
					wallThickness -= 1
				elif (wallThickness < 2):
					wallThickness += 1
		# Right
		if (maze[mazeX][mazeY][4] == 1):
			wallThickness = 2
			range1 = xrange(globals.gRoomH-4)
			range2 = xrange(globals.gRoomH-4, globals.gRoomH)
			for y in range1:
				for x in xrange(wallThickness):
					self.grid[(globals.gRoomW - 1) - x][y] = 1
				wallThickness = self.__modifyWallWidth(wallThickness)
			for y in range2:
				for x in xrange(wallThickness):
					self.grid[(globals.gRoomW - 1) - x][y] = 1
				if (wallThickness > 2):
					wallThickness -= 1
				elif (wallThickness < 2):
					wallThickness += 1
		# Always fill in corners, taking into account widths of adjacent walls
		self.grid[0][0] = 1
		self.grid[0][1] = 1
		self.grid[1][1] = 1
		self.grid[1][0] = 1
		self.grid[globals.gRoomW - 1][0] = 1
		self.grid[globals.gRoomW - 1][1] = 1
		self.grid[globals.gRoomW - 2][1] = 1
		self.grid[globals.gRoomW - 2][0] = 1
		self.grid[globals.gRoomW - 1][globals.gRoomH - 1] = 1
		self.grid[globals.gRoomW - 1][globals.gRoomH - 2] = 1
		self.grid[globals.gRoomW - 2][globals.gRoomH - 2] = 1
		self.grid[globals.gRoomW - 2][globals.gRoomH - 1] = 1
		self.grid[0][globals.gRoomH - 1] = 1
		self.grid[0][globals.gRoomH - 2] = 1
		self.grid[1][globals.gRoomH - 2] = 1
		self.grid[1][globals.gRoomH - 1] = 1
		# In starting room, always ensure space for player to take off
		if (startRoom):
			for y in xrange(1, 5):
				for x in xrange(-1, 2):
					self.grid[(globals.gRoomW/2) + x][y] = 0
		# In Adventure game mode, create enemies!
		if (globals.gGameMode == 0):
			self.enemies = pygame.sprite.RenderUpdates()
			self.mobile_enemies = pygame.sprite.RenderUpdates()
			self.boss_enemies = pygame.sprite.RenderUpdates()
			# Create bosses for keycard rooms
			if (self.bossRoom):
				if (self.whichKey == 0):
					spEnemies.Lobster(globals, player, self.boss_enemies, bulletgroup, self)
				if (self.whichKey == 1):
					spEnemies.Mango(globals, player, self.boss_enemies, bulletgroup, self)
				# TODO: Third boss
				# Cannot leave a boss room during battle - put borders all around
				for x in xrange(0, globals.gRoomW):
					self.grid[x][globals.gRoomH - 1] += 6
				for x in xrange(0, globals.gRoomW):
					self.grid[x][0] += 6
				for y in xrange(1, globals.gRoomH - 1):
					self.grid[globals.gRoomW - 1][y] += 6
				for y in xrange(1, globals.gRoomH - 1):
					self.grid[0][y] += 6
				return
			if (not startRoom):
				# Random obstacles
				# Key on difficulty based on keys retrieved so far
				for i in xrange(random.randint(0, len(player.heldKeys * 3) + 1)):
					x = random.randint(3, globals.gRoomW - 4)
					y = random.randint(3, globals.gRoomH - 4)
					self.grid[x][y] = 1
			self.enemies = pygame.sprite.RenderUpdates()
			self.mobile_enemies = pygame.sprite.RenderUpdates()
			# Mines
			# Key on difficulty
			for i in xrange(random.randint(0, len(player.heldKeys * 2) + 1)):
				spEnemies.Mine(globals, self.mobile_enemies)
			# Turrets & lightning rods
			# Create 0 to 10 turrets
			# Make dependent on current difficulty
			for i in xrange(random.randint(0, len(player.heldKeys * 3) + 1)):
				# Pick random position, but not starting row/column
				if (random.randint(0, 1)):
					# Put facing left/right on a chosen row
					if (startRoom):
						y = random.randint(6, globals.gRoomH - 2)
					else:
						y = random.randint(1, globals.gRoomH - 2)
					# Start seeking from left or right
					if (random.randint(0, 1) == 1):
						myrange = xrange(1, globals.gRoomW - 1)
					else:
						myrange = xrange(globals.gRoomW - 2, 0, -1)
					for j in myrange:
						# Pick an empty square next to some ground
						if (self.grid[j][y] == 0):
							if (self.grid[j - 1][y] == 1):
								# Put turret facing left to right
								if (random.randint(0, 5) == 5):
									enemy = spEnemies.Laser(j, y, 1, globals, self.enemies, player, self)
								else:
									enemy = spEnemies.Turret(j, y, 1, globals, self.enemies, bulletgroup, player, self)
								self.grid[j][y] = enemy
								break
							elif (self.grid[j + 1][y] == 1):
								# Put turret facing right to left
								if (random.randint(0, 5) == 5):
									enemy = spEnemies.Laser(j, y, 3, globals, self.enemies, player, self)
								else:
									enemy = spEnemies.Turret(j, y, 3, globals, self.enemies, bulletgroup, player, self)
								self.grid[j][y] = enemy
								break
				else:
					# Put facing up/down on a given column
					if (startRoom):
						x = random.randint(1, globals.gRoomW - 5)
						if (x >= ((globals.gRoomW/2) - 1)):
							x += 3
					else:
						x = random.randint(1, globals.gRoomW - 2)
					# Start seeking from top/bottom
					if (random.randint(0,1) == 1):
						myrange = xrange(1, globals.gRoomH - 1)
					else:
						myrange = xrange(globals.gRoomH - 2, 0, -1)
					for j in myrange:
						# Pick an empty square next to some ground
						if (self.grid[x][j] == 0):
							if (self.grid[x][j - 1] == 1):
								# Put turret facing bottom to top
								if (random.randint(0, 5) == 5):
									enemy = spEnemies.Laser(x, j, 0, globals, self.enemies, player, self)
								else:
									enemy = spEnemies.Turret(x, j, 0, globals, self.enemies, bulletgroup, player, self)
								self.grid[x][j] = enemy
								break
							elif (self.grid[x][j + 1] == 1):
								# Put turret facing top to bottom
								if (random.randint(0, 5) == 5):
									enemy = spEnemies.Laser(x, j, 2, globals, self.enemies, player, self)
								else:
									enemy = spEnemies.Turret(x, j, 2, globals, self.enemies, bulletgroup, player, self)
								self.grid[x][j] = enemy
								break

	def __modifyWallWidth(self, wallThickness):
		if (self.bossRoom):
			return wallThickness
		if (random.randint(0,1) == 1):
			wallThickness += 1
		if (random.randint(0,1) == 1):
			wallThickness -= 1
		if (wallThickness < 1):
			wallThickness = 1
		elif (wallThickness > 5):
			wallThickness = 5
		return wallThickness
	
	def drawBackground(self, globals):
		background = spResources.getGraphic(globals.gDatadir, 'background', False)
		hud = spResources.getGraphic(globals.gDatadir, 'hud', False)
		tile = spResources.getGraphic(globals.gDatadir, 'tile', False)
		skull = spResources.getGraphic(globals.gDatadir, 'skull', False)
		surf = pygame.Surface((globals.gXres, globals.gYres))
		surf.convert()
		surf.blit(hud, (0, 0))
		x = 0
		gx = 0
		for column in self.grid:
			y = globals.gYres - globals.gTileH
			gy = 0
			for row in column:
				# Remove borders from now-defunct boss rooms
				if (isinstance(row, int) and row >= 6 and not self.bossRoom):
					row -= 6
					# Remove from the original grid (loop values are values, not references)
					self.grid[gx][gy] -= 6
				if (row == 1):
					surf.blit(tile, (x, y))
				elif (row == 2):
					if (0 in self.__player.heldKeys):
						self.grid[gx][gy] = 0
						row = 0
						surf.blit(background, (x, y))
					else:
						door = spResources.getGraphic(globals.gDatadir, 'greendoor', False)
						surf.blit(door, (x, y))
				elif (row == 3):
					if (1 in self.__player.heldKeys):
						self.grid[gx][gy] = 0
						row = 0
						surf.blit(background, (x, y))
					else:
						door = spResources.getGraphic(globals.gDatadir, 'bluedoor', False)
						surf.blit(door, (x, y))
				elif (row == 4):
					if (2 in self.__player.heldKeys):
						self.grid[gx][gy] = 0
						row = 0
						surf.blit(background, (x, y))
					else:
						door = spResources.getGraphic(globals.gDatadir, 'reddoor', False)
						surf.blit(door, (x, y))
				elif (isinstance(row, int) and row >= 6):
					surf.blit(skull, (x, y))
				else:
					surf.blit(background, (x, y))
				y -= globals.gTileH
				gy += 1
			x += globals.gTileW
			gx += 1
		return surf
	
	def squareIsSolid(self, x, y):
		return (self.grid[x][y] > 0)
